/*
 *  netrik -- The ANTRIK Internet Viewer
 *  Copyright (C) Olaf D. Buddenhagen AKA antrik, et al (see AUTHORS)
 *  Published under the GNU GPL; see LICENSE for details.
 */
/*
 * http-parse-header.c -- parse the http header
 *
 * (C) 2002 Patrice Neff
 *     2002 antrik
 */
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "debug.h"
#include "load.h"
#include "http-parse-header.h"

static struct Http_handle *handle;

static void error(char*);
static void header(void);

static void statusline(void);
static void version(void);
static void statuscode(void);
static void statusword(void);

static int in_dataline(void);
static void dataline(void);
static void name(void);
static int in_token(void);
static void token(void);
static void datatoken(void);	/* the same as token(), but stores the value in the current header */
static int in_nametoken(void);
static void nametoken(void);
static void data(void);
static int in_folding(void);
static void folding(void);

static int in_space(void);
static void space(void);
static void newline(void);
static int in_number(void);
static void number(void);
static int in_letter(void);
static void letter(void);

static int load_buffer(void);    /* refill read buffer if empty */
static char sgetc(void);	/* get one character from the socket */
static char FOLLOW(void);  /* returns the following character, but doesn't remove it */

#ifdef DEBUG
static void dump_headers(struct Header_list *headers);    /* dump all extracted headers */
#endif

static void error(char* msg) {
   fprintf(stderr, "HTTP header parsing error (%s)\n", msg);
   exit(2);
}

static void header(void) {
   DMSG(("header "));
   statusline();
   newline();
   dataline();
   while(in_dataline())
      dataline();
   newline();
}

static void statusline(void) {
   DMSG(("statusline "));
   /* first word: "HTTP/" */
   if(sgetc()!='H' || sgetc()!='T' || sgetc()!='T' || sgetc()!='P' || sgetc()!='/')
      error("HTTP/ expected");

   version();
   space();
   statuscode();
   space();
   statusword();
}

static void version(void) {
   char tmp[4];
   int i;

   DMSG(("version "));

   for(i=0; i<3; ++i)
      tmp[i] = sgetc();
   tmp[3] = '\0';

   if(strncmp(tmp, "0.9", 4)!=0 &&
      strncmp(tmp, "1.0", 4)!=0 &&
      strncmp(tmp, "1.1", 4)!=0)
      error("version expected");
}

static void statuscode(void) {
   DMSG(("statuscode "));

   number();
   number();
   number();
}

static void statusword(void) {
   DMSG(("statusword "));

   letter();
   while(in_token())
      token();
}

static int in_dataline(void) {
   return(in_nametoken());
}

static void dataline(void) {
   struct Header_list*	hl;    /* new header entry */
   
   hl = malloc(sizeof(struct Header_list));
   if(hl==NULL) {
      fprintf(stderr, "memory allocation error while parsing headers (in function dataline)\n");
      exit(1);
   }

   hl->next = NULL;
   hl->name = NULL;
   hl->value = NULL;
   if(handle->headers == NULL)
      handle->headers = hl;
   else {
      struct Header_list* i = handle->headers;
      while(i->next != NULL) i = i->next;
      i->next = hl;
   }
   handle->last_header = hl;
	
   DMSG(("dataline "));

   name();

   /* colon */
   if(sgetc()!=':')
      error(": expected");

   space();
   data();
   newline();

   while(in_folding())
      folding();
}

static void name(void) {
   DMSG(("name "));

   nametoken();
   while(in_nametoken())
      nametoken();
}

static int in_nametoken(void) {
   return(in_letter() || in_number() || FOLLOW()=='-');
}

static void nametoken(void) {
   int i;
   DMSG(("nametoken "));

   if(!in_nametoken())
      error("letter, number or dash expected");

   if(handle->last_header->name!=NULL)
      i = strlen(handle->last_header->name);
   else
      i = 0;
   handle->last_header->name = realloc(handle->last_header->name, i+2);
   if(handle->last_header->name==NULL) {
      fprintf(stderr, "memory allocation error while parsing headers (in function nametoken)\n");
      exit(1);
   }
   handle->last_header->name[i] = sgetc();
   handle->last_header->name[i+1] = '\0';
}

static void data(void) {
   DMSG(("data "));

   datatoken();
   while(in_token())
      datatoken();
}

static int in_token(void) {
   char c = FOLLOW();
   return(c!='\r' && c!='\n');
}

static void token(void) {
   DMSG(("token "));

   if(!in_token())
      error("token expected");
   sgetc();
}

static void datatoken(void) {
   int i;
   DMSG(("datatoken "));

   if(!in_token())
      error("datatoken expected");

   if(handle->last_header->value!=NULL)
      i = strlen(handle->last_header->value);
   else
      i = 0;
   handle->last_header->value = realloc(handle->last_header->value, i+2);
   if(handle->last_header->value==NULL) {
      fprintf(stderr, "memory allocation error while parsing headers (in function datatoken)\n");
      exit(1);
   }
   handle->last_header->value[i] = sgetc();
   handle->last_header->value[i+1] = '\0';
}

static int in_folding(void) {
   return(FOLLOW()==' ');
}

static void folding(void) {
   int i;
   DMSG(("folding "));

   space();
   while(in_space())
      space();

   /* add *one* space to the header value */
   if(handle->last_header->value!=NULL)
      i = strlen(handle->last_header->value);
   else
      i = 0;
   handle->last_header->value = realloc(handle->last_header->value, i+2);
   if(handle->last_header->value==NULL) {
      fprintf(stderr, "memory allocation error while parsing headers (in function folding)\n");
      exit(1);
   }
   handle->last_header->value[i] = ' ';
   handle->last_header->value[i+1] = '\0';

   data();
   newline();
}

static int in_space(void) {
   char c = FOLLOW();

   return(c==' ' || c=='\t');
}

static void space(void) {
   DMSG(("space "));

   if(!in_space())
      error("space expected");
   sgetc();
}

static void newline(void) {
   DMSG(("newline "));

   if(sgetc()!='\r' || sgetc()!='\n') {
      error("newline expected");
   }
}

/* antrik --> */
static int in_number(void) {
   return(isdigit(FOLLOW()));
}
/* <-- antrik */

static void number(void) {
   DMSG(("number "));

   if(!isdigit(sgetc()))
      error("number expected");
}

static int in_letter(void) {
   return(isalpha(FOLLOW()));
}

static void letter(void) {
   DMSG(("letter "));

   if(!in_letter())
      error("character expected");
   sgetc();
}

/* antrik --> */
static struct Resource	*input;

/* refill read buffer if empty */
static int load_buffer(void)
{
   if(input->buf_ptr==input->buf_end) {    /* buf empty */
      load(input);    /* -> read next block */
      input->buf_ptr=input->buf;    /* reset read pointer */
      if(input->buf_ptr==input->buf_end)    /* still empty */
	 return -1;    /* EOF */
      else
	 return 1;    /* refilled */
   } else    /* buf not empty */
      return 0;    /* do nothing */
}

/* read a character from socket */
static char sgetc(void)
{
   /* fill buffer */
   switch(load_buffer()) {
      case 1:    /* refilled */
	 DMSG(("\n\"%*s\"\n", input->buf_end-input->buf_ptr, input->buf_ptr));
	 break;
      case -1:    /* couldn't refill (EOF) */
	 fprintf(stderr, "unexpected end of file\n");
	 exit(2);
      case 0:    /* didn't refill (wasn't empty) */
	 break;
   }

   /* read char from buffer */
   return *input->buf_ptr++;
}

/* read a character from socket, but don't advance read pointer */
static char FOLLOW(void)
{
   char in;

   in=sgetc();    /* read char */
   --input->buf_ptr;    /* keep read position (undo advance) */
   return in;
}

#ifdef DEBUG
/* dump all extracted headers */
void dump_headers(headers)
struct Header_list	*headers;
{
   debug_printf("\nheaders:\n");

   for(; headers!=NULL; headers=headers->next)
      debug_printf("%s: %s\n", headers->name, headers->value);

   debug_printf("\n");
}
#endif
/* <-- antrik */

/* Patrice, antrik --> */
void parse_header(struct Resource *res) {
   DMSG(("parsing HTTP headers...\n"));
   handle=res->handle.http;
   input=res;

   if(load_buffer()==-1) {    /* EOF before anything read */
      fprintf(stderr, "no data\n");
      return;
   }
   DMSG(("\n\"%*s\"\n", input->buf_end-input->buf_ptr, input->buf_ptr));

   header();
#ifdef DEBUG
   if(cfg.debug)
      dump_headers(handle->headers);
#endif
}
/* <-- Patrice, antrik */
