/*
   netrik -- The ANTRIK Internet Viewer
   Copyright (C) Olaf D. Buddenhagen AKA antrik, et al (see AUTHORS)
   Published under the GNU GPL; see LICENSE for details.
*/
/*
 * http.c -- load files via http
 *
 * (C) 2001, 2002 Patrice Neff
 *     2001, 2002 antrik
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

/* network includes */
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>

#include "debug.h"
#include "forms.h"    /* to get form data */
#include "http.h"
#include "http-parse-header.h"
#include "url.h"

static int get_http_socket(const struct Url *url, const struct Item *form);    /* open connection to the server */
static char *str_realloc_clear(char *ptr, int inc);
static char *get_http_cmd(const struct Url *url, const struct Url *proxy, const struct Item *form);    /* construct a HTTP command */
static char *http_encode(const char* url);	/* convert illegal URL characters */

/* reallocate memory, frees the allocated memory if
   this is the first allocation */
static char *str_realloc_clear(ptr, inc)
char *ptr;
int inc;
{
   char *mem=0;
   int fresh=0;
   int prevlen=0;

   if(ptr==NULL) {
      prevlen=0;
      fresh=1;
   } else {
      prevlen=strlen(ptr);
      fresh=0;
   }

   mem=(char *)realloc(ptr, prevlen+inc);
   if(mem==NULL) {
      fprintf(stderr, "memory allocation failure\n");
      exit(1);
   }
   if(fresh)
      memset(mem, 0, inc);
   else
      memset(mem+prevlen, 0, inc);

   return(mem);
}

/* Patrice, antrik --> */
static char *http_encode(url)
const char* url;
{
   int i=0;
   int j=0;
   int len=0;
   char *nurl=NULL;

   while(url[i]!='\0') {
      switch(url[i]) {
         case ' ': len+=3; break;
         default:  ++len; break;
      }
      ++i;
   }

   i=0;
   nurl=str_realloc_clear(nurl, len+1);
   nurl[len]='\0';
   while(url[i]!='\0') {
      switch(url[i]) {
         case ' ': strcat(nurl, "%20"); j+=3; break;
         default:  nurl[j++]=url[i]; break;
      }
      ++i;
   }
   return(nurl);
}

/* construct HTTP GET command */
static char *get_http_cmd(url, proxy, form)
const struct Url	*url;
const struct Url	*proxy;
const struct Item	*form;
{
   char	*cmd=NULL;
   int	len;

   /* calculate length of HTTP request */
   len=strlen("GET ");
   if(form!=NULL) {    /* submit form data */
      struct Form_handle	handle=form_start(form, 1);    /* state var for iterating through form elements (only succesful ones) */
      struct Link		*link;    /* link struct of current form element */

      switch(url->path[strlen(url->path)-1]) {    /* last URL char */
         case '?': case '&':    /* can directly append form parameters */
	    break;
         default:    /* need to put '&' or '?' */
	    ++len;
	    break;
      }

      while((link=form_next(&handle)) != NULL) {    /* for all successful form (data) elements */
	 char *tmp;

	 len+=strlen(tmp=http_encode(link->name));
	 free(tmp);
	 len+=strlen(tmp=http_encode(link->value));
	 free(tmp);
	 len+=2;    /* equal sign + ampersand */
      }
   }
   len+= proxy!=NULL ? strlen(url->full_url) : strlen(url->path);
   len+=strlen(" HTTP/1.1\r\nHost: ");
   len+=strlen(url->host);
   len+=strlen("\r\nConnection: Close\r\n\r\n");

   /* build request */
   cmd=str_realloc_clear(cmd, len+1);
   strcpy(cmd, "GET ");
   if(proxy==NULL)
      strcat(cmd, url->path);
   else
      strcat(cmd, url->full_url);
   if(form!=NULL) {    /* submit form data */
      struct Form_handle	handle=form_start(form, 1);    /* state var for iterating through form elements (only succesful ones) */
      struct Link		*link;    /* link struct of current form element */

      switch(url->path[strlen(url->path)-1]) {    /* last URL char */
         case '?': case '&':    /* can directly append form parameters */
	    break;
         default:    /* need to put '&' or '?' */
	    if(strchr(url->path, '?'))    /* already has parameters -> put separator */
	       strcat(cmd, "&");
	    else    /* no parameters yet */
	       strcat(cmd, "?");
	    break;
      }

      while((link=form_next(&handle)) != NULL) {    /* for all successful form (data) elements */
	 char *tmp;

	 strcat(cmd, tmp=http_encode(link->name));
	 free(tmp);
	 strcat(cmd, "=");
	 strcat(cmd, tmp=http_encode(link->value));
	 free(tmp);
	 strcat(cmd, "&");
      }
      *(strchr(cmd, '\0')-1)='\0';    /* remove '&' at end */
   }
   strcat(cmd, " HTTP/1.1\r\nHost: ");
   strcat(cmd, url->host);
   strcat(cmd, "\r\nConnection: Close\r\n\r\n");

   return cmd;
}

/* <-- Patrice, antrik */

/*
 * Prepare http connection for reading data.
 * 
 * Creates an HTTP handle, opens a socket, connects to the server, sends HTTP
 * request (optionally submitting form data), loads and parses the response
 * HTTP header.
 */
void http_init_load(res, form)
struct Resource		*res;
const struct Item	*form;
{
   /* alloc http struct */
   res->handle.http=malloc(sizeof(struct Http_handle));
   if(res->handle.http==NULL) {
      fprintf(stderr, "memory allocation failure while opening resource\n");
      exit(1);
   }

   res->handle.http->headers=NULL;
   res->handle.http->last_header=NULL;
   res->handle.http->socket=get_http_socket(res->url, form);
   if(res->handle.http->socket!=-1) {    /* could open HTTP connection */
      res->type=RES_HTTP;
      parse_header(res);
   } else {    /* HTTP connection failed */
      res->type=RES_FAIL;
      res->url->proto.type=PT_INTERNAL;    /* don't keep in history */
   }
}

/* Patrice, antrik --> */

/* open a connection to the http server */
static int get_http_socket(url, form)
const struct Url	*url;
const struct Item	*form;
{
   struct hostent *hp;		/* host info of the target host */
   struct sockaddr_in sap;	/* connect information */
   int sock;			/* the socket */
   char *http_command;		/* http command to be sent to the server */
   char *proxy;
   struct Url *proxy_url;
   char *connect_host;    /* server connected to get the page (proxy when used, target server otherwise) */
   int connect_port;

   /* get url information */
   if(cfg.proxy) {
      proxy=getenv("http_proxy");
      if(proxy==NULL)
	 proxy=getenv("HTTP_PROXY");
   } else
      proxy=NULL;

   if(proxy!=NULL) {
      DMSG(("parsing proxy URL...\n"));
      proxy_url=split_url(proxy);
      if(proxy_url->proto.type==PT_INTERNAL) {    /* splitting failed */
	 fprintf(stderr, "can't parse proxy URL\n");
	 return -1;
      }
   } else
      proxy_url=NULL;

   if(proxy==NULL) {    /* no proxy */
      connect_host=url->host;
      connect_port=url->port;
   } else {    /* use proxy */
      connect_host=proxy_url->host;
      connect_port=proxy_url->port;
   }
   if(connect_port==0)
      connect_port=80;

   /* start to fill the sockaddr struct */
   memset(&sap, 0, sizeof(sap));
   sap.sin_family = AF_INET;
   sap.sin_port = htons(connect_port);

   /* address to connect to */
   if(connect_host!=NULL) {
      if(!inet_aton(connect_host, &sap.sin_addr)) {
         hp = gethostbyname(connect_host);
         if(hp==NULL) {
            fprintf(stderr, "unknown host: %s\n", connect_host);
	    return -1;
         }
         sap.sin_addr = *(struct in_addr*)hp->h_addr;
      }
   } else {
      sap.sin_addr.s_addr = htonl(INADDR_ANY);
   }

   /* connect to the server */
   DMSG(("connecting to %s:%i\n", connect_host, connect_port));
   sock = socket(AF_INET, SOCK_STREAM, 0);
   if(connect(sock, (struct sockaddr*)&sap, sizeof(sap))) {
      fprintf(stderr, "connect to host %s failed\n", connect_host);
      return -1;
   }

   /* build the http commands */
   http_command=get_http_cmd(url, proxy_url, form);
   DMSG(("submitting HTTP command:\n%s\n", http_command));

   /* send the http commands to fetch the URL */
   if(write(sock, http_command, strlen(http_command))==-1) {
      fprintf(stderr, "sending HTTP request failed\n");
      return -1;
   }

   /* free memory */
   free(http_command);
   if(proxy_url!=NULL)
      free_url(proxy_url);

   return sock;
}

/* <-- Patrice, antrik */
