/*
   netrik -- The ANTRIK Internet Viewer
   Copyright (C) Olaf D. Buddenhagen AKA antrik, et al (see AUTHORS)
   Published under the GNU GPL; see LICENSE for details.
*/
/*
 * pager.c -- Interactive viewing module.
 *
 * (C) 2001 antrik
 *
 * Displays output page and handles user commands.
 */
#include <curses.h>
#include <stdio.h>
#include <stdlib.h>

#include "cfg.h"
#include "items.h"    /* for TM_ACTIVE */
#include "layout.h"
#include "links.h"
#include "pager.h"
#include "render.h"
#include "screen.h"
#include "url-history.h"

static void activate_link(struct Page *page, int change_link);    /* set active link */
static void scroll_to(struct Page *page, int new_line);    /* adjust visible part of page (return new first visible line) */
static void activate_anchor(struct Page *page);    /* show active anchor */

#define PAGE_INVALID ((unsigned int)-1>>1)    /* biggest possible positive int */
static int	old_line;    /* page position visible up to now */

/* set active link;
 * also highlights it and redraws the affected screen area.
 * if link_num is -1, deactivate active link */
static void activate_link(page, change_link)
struct Page	*page;
int		change_link;
{
   const int			link_num=change_link>=0?change_link:page->active_link;    /* change given link; if -1, change old one (deactivate) */

   const struct Link_ptr	*list_entry=&page->links->link[link_num];    /* entry in link list pointing to link */
   const struct Item		*item=list_entry->item;    /* item containing link */
   const struct String		*string=item->data.string;    /* string containing link */
   const struct Link		*link=&string->link[list_entry->num];    /* link data */

   int	x_start, y_start, x_end, y_end;    /* screen area affected by link */
   int	page_x, page_y;

   highlight_link(page, link_num, change_link==-1?:0);

   /* screen coordinates of updated area */
   if(link->y_end-link->y_start==1) {    /* link on one line */
      page_x=link->x_start;
      x_start=link->x_start;    /* no x-scrolling yet... */
      x_end=link->x_end;
   } else {    /* link spans multiple lines */
      page_x=item->x_start;
      x_start=item->x_start;
      x_end=item->x_end;
   }
   page_y=link->y_start;
   y_start=link->y_start-page->pager_pos;
   y_end=link->y_end-page->pager_pos;

   if(x_start<COLS && x_end>0 && y_start<LINES && y_end>0) {    /* some part visible */
      /* truncate to visible area */
      if(x_start<0) {
	 page_x-=x_start;
	 x_start=0;
      }
      if(y_start<0) {
	 page_y-=y_start;
	 y_start=0;
      }
      if(x_end>COLS)
	 x_end=COLS;
      if(y_end>LINES)
	 y_end=LINES;

      render(page, page_x, page_y, x_start, y_start, x_end-x_start, y_end-y_start, 0);    /* update area affected by link */
   }

   page->active_link=change_link;
}

/* move visible page area;
 * return new first visible line */
static void scroll_to(page, new_line)
struct Page	*page;
int		new_line;    /* page position to move to */
{
   int		scroll_lines;    /* number of lines to scroll */

   /* page boundaries */
   if(new_line>page->item_tree->y_end-LINES)    /* page ends before end of screen -> adjust */
      new_line=page->item_tree->y_end-LINES;
   if(new_line<0)    /* page begins after start of screen -> adjust (may undo previous condition, if page length < screen length!) */
      new_line=0;

   scroll_lines=new_line-old_line;

   if(abs(scroll_lines)<=LINES) {    /* move not more than a screenfull -> scroll */
      move(0,0);
      if(insdelln(-scroll_lines)==ERR) {
	 endwin();
	 fprintf(stderr, "curses error: can't scroll (in function scroll_to)\n");
	 exit(2);
      }
      
      if(scroll_lines<0)    /* scroll back */
	 render(page, 0, new_line, 0, 0, COLS, -scroll_lines, 0);    /* render new first line(s) */
      else    /* scroll forward */
	 render(page, 0, new_line+LINES-scroll_lines, 0, LINES-scroll_lines, COLS, scroll_lines, 0);    /* render new last line(s) */

   } else {    /* move more than screenful (or PAGE_INVALID) -> repaint */
      erase();
      render(page, 0, new_line, 0, 0, COLS, LINES, 0);    /* fill whole screen */
   }
   
   old_line=new_line;    /* save new position */
   page->pager_pos=new_line;

   if(page->active_link>=0) {    /* some link is active */
      const struct Link_ptr	*list_entry=&page->links->link[page->active_link];    /* entry in link list pointing to current link */
      const struct Link		*link=&list_entry->item->data.string->link[list_entry->num];    /* current link's data */

      if((link->y_start-page->pager_pos < cfg.link_margin    /* link start inside upper "link_margin" */
          && page->pager_pos>0)    /* allow anyway if at page top */
         || (link->y_end-page->pager_pos > LINES-cfg.link_margin    /* link end inside lower "link_margin" */
	     && page->pager_pos < page->item_tree->y_end-LINES)    /* allow anyway if at page bottom */
	)    /* moved out of legal area for active links */
	 activate_link(page, -1);    /* -> deactivate */
   }
}

/*
 * Show active anchor.
 *
 * Scroll so that the anchor will be in the desired screen line (determined by
 * "cfg.anchor_offset"); draw marks at all screen lines spanned by the anchor.
 * When called while no "active_anchor" set, clear anchor marks.
 *
 * First the anchor position is retrieved and the page scrolled to the optimal
 * position. Afterwards, the real anchor position on the screen is calculated,
 * and the marks set directly to the screen. (Not added to "item_tree".)
 * Clearing is done by rerendering the affected area.
 */

static void activate_anchor(page)
struct Page	*page;
{
   static int	anchor;    /* affected anchor */
   int		start_line;    /* screen line containing anchor start */
   int		end_line;    /* screen lines after anchor end */

   if(page->active_anchor>=0)    /* there is an anchor to activate -> take it (otherwise use the one stored before) */
      anchor=page->active_anchor;

   /* get screen position of anchor */
   {
      const struct Item	*anchor_item=page->anchors->anchor_item[anchor];

      if(page->active_anchor>=0)    /* activating anchor -> first need to scroll */
	 scroll_to(page, anchor_item->y_start-(cfg.anchor_offset?LINES/cfg.anchor_offset:cfg.link_margin));    /* scroll so anchor will appear "1/anchor_offset" below screen top, or "link_margin" below screen top if no "anchor_offset" */

      start_line=anchor_item->y_start - page->pager_pos;
      end_line=anchor_item->y_end - page->pager_pos;

      if(end_line==start_line)    /* empty anchor -> mark one line */
	 ++end_line;

      if(start_line<0)    /* starts before screen top -> truncate */
	 start_line=0;

      if(end_line>LINES)
	 end_line=LINES;
   }

   if(page->active_anchor>=0) {    /* activate -> draw marks */
      int	line;

#ifdef DEBUG
      if(start_line>=LINES || end_line<=0) {
	 endwin();
	 fprintf(stderr, "internal error: trying to activate anchor not on screen\n");
	 exit(100);
      }
#endif

      for(line=start_line; line<end_line; ++line) {
	 set_color(color_map[TM_SYS]);
	 mvprintw(line, COLS-1, "%c", '<');
      }
   } else    /* deactivate -> rerender area containing marks */
      if(end_line>start_line)    /* anything to render (not scrolled out of screen) */
	 render(page, COLS-1, start_line+page->pager_pos, COLS-1, start_line, 1, end_line-start_line, 1);    /* overpaint */
}

/*
 * Interactive pager.
 *
 * Until some command requirering returning from pager (following link,
 * pressing 'q' etc.) issued: Display part of output page on curses screen;
 * read input key; move visible part according to pressed key function.
 *
 * When some command was issued that requires quitting the pager and returning
 * to the main program, return a flag indicating what was the cause, so main()
 * knows what to do.
 */

enum Pager_ret display(page)
struct Page	*page;
{
   enum Pager_ret	quit=RET_NO;

   /* create initial pager screen */
   if(cbreak()==ERR) {    /* readline resets this... */
      endwin();
      fprintf(stderr, "curses error while (re)starting pager");
      exit(2);
   }

   old_line=PAGE_INVALID;    /* force repaint */

   if(page->active_anchor>=0)    /* anchor requested -> jump to it */
      activate_anchor(page);
   else    /* no anchor -> just go to current page position */
      scroll_to(page, page->pager_pos);    /* display visible part of page */

   if(page->active_link>=0)    /* some link active -> force highlighting (necessary after reloading from history) */
      activate_link(page, page->active_link);

   do {    /* until quit */
      int	cursor_x, cursor_y;

      if(page->active_link >= 0) {
	 const struct Link_ptr	*list_entry=&page->links->link[page->active_link];    /* entry in link list pointing to link */
	 const struct Item	*item=list_entry->item;    /* item containing link */
	 const struct Link	*link=&item->data.string->link[list_entry->num];    /* link's data */

	 cursor_x=link->x_start;
	 cursor_y=link->y_start-page->pager_pos;
      } else
	 cursor_x=cursor_y=0;

#ifdef DEBUG
      if((unsigned)cursor_x>=(unsigned)COLS || (unsigned)cursor_y>=(unsigned)LINES) {
	 endwin();
	 fprintf(stderr, "internal error: trying to set cursor outside screen (x: %d; y: %d)\n", cursor_x, cursor_y);
	 exit(100);
      }
#endif

      switch(mvgetch(cursor_y, cursor_x)) {

	 /* screen backward */
	 case 'B'-'@':    /* ^b */
	 case '-':
	    scroll_to(page, page->pager_pos-LINES);
	    break;

	 /* half screen forward */
	 case 'D'-'@':    /* ^d */
	 case KEY_NPAGE:
	    scroll_to(page, page->pager_pos+LINES/2);
	    break;

	 /* screen forward */
	 case 'F'-'@':    /* ^f */
	 case ' ':
	    scroll_to(page, page->pager_pos+LINES);
	    break;

	 /* reload page */
	 case 'R'-'@':    /* ^r */
	    if(page->url->proto.type!=PT_INTERNAL)    /* normal page */
	       quit=RET_HISTORY;
	    else    /* stdin -> can't reload */
	       flash();
	    break;

	 /* half screen backward */
	 case 'U'-'@':    /* ^u */
	 case KEY_PPAGE:
	    scroll_to(page, page->pager_pos-LINES/2);
	    break;

	 /* enter one command in ex mode */
	 case ':':
	    quit=RET_COMMAND;
	    break;

	 /* back to previous URL */
	 case 'b':
	 case KEY_LEFT:
	    if(page_list.pos > 0) {    /* any URL(s) to go back */
	       --page_list.pos;
	       quit=RET_HISTORY;
	    } else
	       flash();
	    break;

	 /* back to URL before last absolute link/load */
	 case 'B':
	    if(page_list.pos > 0) {    /* any URL(s) to go back */
	       int	pos;

	       for(pos=page_list.pos-1; pos>0; --pos)    /* traverse history backwards (stop at first entry) */
		  if(page_list.page[pos+1]->url->absolute)    /* found matching entry (a page before one with absolute URL) */
		     break;
	       page_list.pos=pos;    /* take found entry (or first entry, if nothing found) */

	       quit=RET_HISTORY;
	    } else    /* already at first entry */
	       flash();
	    break;

	 /* show current page URL */
	 case 'c':
	    quit=RET_URL;
	    break;

	 /* forward to next URL (after 'b' or other commands going back in history) */
	 case 'f':
	 case KEY_RIGHT:
	    if(page_list.pos < page_list.count-1) {    /* can go forward (not at last entry) */
	       ++page_list.pos;
	       quit=RET_HISTORY;
	    } else
	       flash();
	    break;

	 /* forward to URL before next absolute link/load */
	 case 'F':
	    if(page_list.pos < page_list.count-1) {    /* can go forward (not at last entry) */
	       int	pos;

	       for(pos=page_list.pos+1; pos<page_list.count-1; ++pos)    /* traverse history forward (stop at last entry) */
		  if(page_list.page[pos+1]->url->absolute)    /* found matching entry (a page before one with absolute URL) */
		     break;
	       page_list.pos=pos;    /* take found entry (or last entry, if nothing found) */

	       quit=RET_HISTORY;
	    } else    /* already at last entry */
	       flash();
	    break;

	 /* page top */
	 case 'g':
	 case KEY_HOME:
	    scroll_to(page, 0);
	    break;

	 /* page bottom */
	 case 'G':
	 case KEY_END:
	    scroll_to(page, page->item_tree->y_end-LINES);
	    break;

	 /* one line down */
	 case 'j':
	    scroll_to(page, page->pager_pos+1);
	    break;

	 /* down/next link */
	 case 'J':
	 case KEY_DOWN: {
	    int	next_link;    /* link to be activated (if no link previously active, this is the first one in/below visible page area) */
	    int	next_link_visible;    /* "next_link" is inside legal area for active links */

	    if(page->active_link>=0) {    /* some link is active -> try activating next one */
	       next_link=page->active_link+1;
	       if(next_link==page->links->count)    /* active link was already last one... */
		  next_link=-1;
	    } else {    /* no link active -> find first link in visible page area (or below) */
	       int	min_y;    /* first screen line in which links may be activated */

	       if(page->pager_pos>0)    /* not at page top */
		  min_y=cfg.link_margin;
	       else    /* page top -> allow also inside "link_margin"  */
		  min_y=0;

	       for(next_link=0; next_link<page->links->count; ++next_link) {    /* all links */
		  const struct Link_ptr	*list_entry=&page->links->link[next_link];    /* entry in link list pointing to link */
		  const struct Link	*link=&list_entry->item->data.string->link[list_entry->num];    /* link's data */

		  if(link->y_start-page->pager_pos >= min_y)    /* found */
		     break;
	       }
	       if(next_link==page->links->count)    /* nothing found */
		  next_link=-1;
	    }    /* no link currently active */

	    next_link_visible=0;    /* assume no link in legal area */
	    if(next_link>=0) {    /* some candidate found -> test it */
	       const struct Link_ptr	*list_entry=&page->links->link[next_link];    /* entry in link list pointing to link */
	       const struct Link	*link=&list_entry->item->data.string->link[list_entry->num];    /* link's data */

	       if(link->y_end-page->pager_pos <= LINES-cfg.link_margin    /* link ends outside lower "link_margin" */
		  || page->pager_pos >= page->item_tree->y_end-LINES    /* always allow if at page bottom */
		 )    /* inside legal area for active links */
		  next_link_visible=1;
	    }

	    if(!next_link_visible) {    /* no more links in current legal area -> scroll one line */
	       scroll_to(page, page->pager_pos+1);

	       /* retest if *now* the candidate is legal */
	       if(next_link>=0) {
		  const struct Link_ptr	*list_entry=&page->links->link[next_link];    /* entry in link list pointing to link */
		  const struct Link	*link=&list_entry->item->data.string->link[list_entry->num];    /* link's data */

		  if(link->y_end-page->pager_pos <= LINES-cfg.link_margin    /* link ends inside lower "link_margin" */
		     || page->pager_pos >= page->item_tree->y_end-LINES    /* always allow if at page bottom */
		    )    /* inside legal area for active links */
		     next_link_visible=1;
	       }
	    }

	    if(next_link_visible) {    /* have a legal new link -> activate it */
	       if(page->active_link>=0)    /* some link is active */
		  activate_link(page, -1);    /* deactivate old link */
	       activate_link(page, next_link);
	    }

	    break;
	 }

	 /* one line up */
	 case 'k':
	 case KEY_BACKSPACE:
	    scroll_to(page, page->pager_pos-1);
	    break;

	 /* up/next link */
	 case 'K':
	 case KEY_UP: {
	    int	prev_link;    /* link to be activated (if no link previously active, this is the first one above visible page area) */
	    int	prev_link_visible;    /* "prev_link" is inside legal area for active links */

	    if(page->active_link>=0)    /* some link is active -> try activating previous one */
	       prev_link=page->active_link-1;
	    else
	       prev_link=-1;

	    prev_link_visible=0;    /* assume no link in legal area */
	    if(prev_link>=0) {    /* some candidate found -> test it */
	       const struct Link_ptr	*list_entry=&page->links->link[prev_link];    /* entry in link list pointing to link */
	       const struct Link	*link=&list_entry->item->data.string->link[list_entry->num];    /* link's data */

	       if(link->y_start-page->pager_pos >= cfg.link_margin    /* link starts outside upper "link_margin" */
		  || page->pager_pos==0    /* always allow if at page top */
		 )    /* inside legal area for active links */
		  prev_link_visible=1;
	    }

	    if(!prev_link_visible) {    /* no more links in current legal area -> scroll one line */
	       scroll_to(page, page->pager_pos-1);

	       if(!prev_link>=0) {    /* no candidate yet -> find last link above visible page area */
		  int	link_num;

		  for(link_num=0; link_num<page->links->count; ++link_num) {    /* all links */
		     const struct Link_ptr	*list_entry=&page->links->link[link_num];    /* entry in link list pointing to link */
		     const struct Link		*link=&list_entry->item->data.string->link[list_entry->num];    /* link's data */

		     if(link->y_start-page->pager_pos <= cfg.link_margin)    /* still inside search area */
			prev_link=link_num;
		     else    /* left search area */
			break;
		  }
	       }    /* no link currently active */

	       /* retest if *now* we have a legal new link */
	       if(prev_link>=0) {
		  const struct Link_ptr	*list_entry=&page->links->link[prev_link];    /* entry in link list pointing to link */
		  const struct Link	*link=&list_entry->item->data.string->link[list_entry->num];    /* link's data */

		  if(link->y_start-page->pager_pos >= cfg.link_margin    /* link starts outside upper "link_margin" */
		     || page->pager_pos==0    /* always allow if at page top */
		    )    /* inside legal area for active links */
		     prev_link_visible=1;
	       }
	    }

	    if(prev_link_visible) {    /* have a legal new link -> activate it */
	       if(page->active_link>=0)    /* some link is active */
		  activate_link(page, -1);    /* deactivate old link */
	       activate_link(page, prev_link);
	    }

	    break;
	 }

	 /* quit */
	 case 'q':
	    quit=RET_QUIT;
	    break;

	 /* return to previous mark */
	 case 'r':
	    if(page_list.pos > 0) {    /* any URL(s) to go back */
	       int	pos;

	       for(pos=page_list.pos-1; pos>0; --pos)    /* traverse history backwards (stop at first entry) */
		  if(page_list.page[pos]->mark)    /* found a marked page */
		     break;
	       page_list.pos=pos;    /* take found entry (or first entry, if nothing found) */

	       quit=RET_HISTORY;
	    } else    /* already at first entry */
	       flash();
	    break;

	 /* forward to next mark */
	 case 'R':
	    if(page_list.pos < page_list.count-1) {    /* can go forward (not at last entry) */
	       int	pos;

	       for(pos=page_list.pos+1; pos<page_list.count-1; ++pos)    /* traverse history forward (stop at last entry) */
		  if(page_list.page[pos]->mark)    /* found a marked page */
		     break;
	       page_list.pos=pos;    /* take found entry (or first entry, if nothing found) */

	       quit=RET_HISTORY;
	    } else    /* already at last entry */
	       flash();
	    break;

	 /* set page mark */
	 case 's':
	    if(page->url->proto.type!=PT_INTERNAL)    /* normal page */
	       page->mark=1;
	    else    /* internal -> can't mark */
	       flash();
	    break;

	 /* show link URL */
	 case 'u':
	    if(page->active_link>=0)    /* some link active */
	       quit=RET_LINK_URL;
	    else
	       flash();
	    break;

	 /* show absolute link target URL */
	 case 'U':
	    if(page->active_link>=0)    /* some link active */
	       quit=RET_ABSOLUTE_URL;
	    else
	       flash();
	    break;

	 /* two lines up */
	 case KEY_IC:    /* INS */
	    scroll_to(page, page->pager_pos-2);
	    break;

	 /* two lines down */
	 case KEY_DC:    /* DEL */
	    scroll_to(page, page->pager_pos+2);
	    break;

	 /* follow active link */
	 case '\r':    /* <return> */
	    if(page->active_link>=0)    /* some link active -> follow */
	       quit=RET_LINK;
	    else
	       flash();
	    break;

#ifdef DEBUG
	 case ERR:
	    endwin();
	    fprintf(stderr, "internal error: getch() failed (can't read keyboard input)\n");
	    exit(100);
	    break;
#endif
	 
	 /* wrong key */
	 default:
	    flash();
      }
      if(page->active_anchor>=0) {
	 page->active_anchor=-1;
	 activate_anchor(page);
      }
   } while(quit==RET_NO);

   end_fullscreen();
   return quit;    /* main program needs to know whether to quit or to enter command prompt */
}
